package jalview.util;

import jalview.datamodel.HiddenMarkovModel;
import jalview.datamodel.SequenceI;
import jalview.io.DataSourceType;
import jalview.io.FileParse;
import jalview.io.HMMFile;
import jalview.io.StockholmFile;
import jalview.schemes.ResidueProperties;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Vector;

/**
 * Processes probability data. The file indexes used in this program represent
 * the index of the location of a family or hmm in their respective files,
 * starting from 0.
 * 
 * @author TZVanaalten
 *
 */
public class HMMProbabilityDistributionAnalyser
{

  Vector<SequenceI> sequences;

  HiddenMarkovModel hmm;

  // contains the raw data produced
  List<ArrayList<Double>> raw = new ArrayList<>();

  // contains binned data
  Map<String, Double> binned = new HashMap<>();

  // location of the family file
  final static String FAMILIES = "C:/Users/TZVanaalten/Pfam-A.full";

  // location of the HMM file
  final static String HMMS = "H:/Desktop/PFAM/HMMs/Pfam-A.hmm";

  // suffix for raw file
  final static String RAW = "/Raw.csv";

  // suffix for binned file
  final static String BINNED = "/Binned.csv";

  // normalisation scale
  final static double SCALE = 100000;

  // current position in file
  int currentFilePosition = 0;

  final static String NL = "\n";

  // current directory
  String currentFolder;

  /**
   * Sets the working directory.
   * 
   * @param path
   */
  public void setFolder(String path)
  {
    currentFolder = path;
  }

  /**
   * Moves a buffered reader to a specific location in the file, delimited by
   * '//'.
   * 
   * @param index
   *          The index of the location in the file.
   * @param br
   * @throws IOException
   */
  public void moveToFile(int index, BufferedReader br) throws IOException
  {
    for (int i = 0; i < index; i++)
    {
      String line = br.readLine();
      while (!"//".equals(line))
      {
        line = br.readLine();
      }
    }

  }
  
  /**
   * Analyses a specified number of families and then saves the data. Before
   * analysing the data, the previous saved data will be imported and after
   * analysing this data is exported back into the file.
   * 
   * @param increments
   *          The number of families to read before saving.
   * @throws IOException
   */
  public void run(int increments) throws IOException
  {

    readPreviousData(currentFolder);

    BufferedReader posReader = new BufferedReader(
            new FileReader(currentFolder + "/CurrentPosition.txt"));
    String line = posReader.readLine();
    posReader.close();
    currentFilePosition = Integer.parseInt(line);

    BufferedReader inputSTO = new BufferedReader(
            new FileReader(FAMILIES));
    BufferedReader inputHMM = new BufferedReader(
            new FileReader(HMMS));

    moveToFile(currentFilePosition, inputHMM);
    moveToFile(currentFilePosition, inputSTO);

    int filesRead = 0;
    while (filesRead < increments)
    {
      FileParse parserSTO = new FileParse(inputSTO, "",
              DataSourceType.FILE);
      readStockholm(parserSTO);

      FileParse parserHMM = new FileParse(inputHMM, "",
              DataSourceType.FILE);
      readHMM(parserHMM);

      if (hmm.getAlphabetType().equals("amino"))
      {
        int count = countValidResidues();
        processData(count);
        filesRead++;
      }
      currentFilePosition++;
    }

    PrintWriter p = new PrintWriter(
            new File(currentFolder + "/CurrentPosition"));
    p.print(currentFilePosition);
    p.close();
    exportData(currentFolder);
    raw.clear();
    binned.clear();

  }

  /**
   * Reads the previous data from both files
   * 
   * @param source
   * @throws IOException
   */
  public void readPreviousData(String source) throws IOException
  {
    readBinned(source);
    readRaw(source);
  }

  /**
   * Reads the previous data from the binned file.
   * 
   * @param source
   * @throws IOException
   */
  public void readBinned(String source) throws IOException
  {
    BufferedReader input = new BufferedReader(
            new FileReader(source + BINNED));
    String line = input.readLine();
    while (!("".equals(line) || line == null))
    {
      binned = new HashMap<>();
      Scanner scanner = new Scanner(line);
      scanner.useDelimiter(",");
      binned.put(scanner.next(), scanner.nextDouble());
      scanner.close();
      line = input.readLine();
    }

    input.close();
  }

  /**
   * Reads the previous data from the raw file.
   * 
   * @param source
   * @throws IOException
   */
  public void readRaw(String source) throws IOException
  {
    BufferedReader input = new BufferedReader(new FileReader(source + RAW));
    String line = input.readLine();
    if (line == null)
    {
      input.close();
      return;
    }
    Scanner numberScanner = new Scanner(line);
    numberScanner.useDelimiter(",");
    raw = new ArrayList<>();
    while (numberScanner.hasNext())
    {
      numberScanner.next();
      raw.add(new ArrayList<Double>());
    }
    numberScanner.close();

    line = input.readLine();
    while (!("".equals(line) || line == null))
    {
      Scanner scanner = new Scanner(line);
      scanner.useDelimiter(",");

      int i = 0;
      while (scanner.hasNext())
      {
        String value;
        value = scanner.next();
        if (!value.equals("EMPTY"))
        {
          raw.get(i).add(Double.parseDouble(value));
        }

        i++;
      }
      scanner.close();
      line = input.readLine();
    }

    input.close();
  }

  /**
   * Counts the number of valid residues in the sequence.
   * 
   * @return
   */
  public int countValidResidues()
  {
    int count = 0;

    for (int width = 0; width < sequences.size(); width++)
    {
      for (int length = 1; length < hmm.getLength(); length++)
      {
        char symbol;
        int alignPos;
        alignPos = hmm.getNodeAlignmentColumn(length);

        symbol = sequences.get(width).getCharAt(alignPos);
        if (ResidueProperties.aminoBackgroundFrequencies
                .containsKey(symbol))
        {
          count++;
        }
      }
    }

    return count;
  }

  /**
   * Processes data, and stores it in both a raw and binned format.
   * 
   * @param count
   */
  public void processData(int count)
  {

    raw.add(new ArrayList<Double>());
    int rawPos = raw.size() - 1;
    for (int width = 0; width < sequences.size(); width++)
    {
      for (int length = 1; length < hmm.getLength(); length++)
      {
        char symbol;
        int alignPos;
        alignPos = hmm.getNodeAlignmentColumn(length);

        symbol = sequences.get(width).getCharAt(alignPos);
        if (ResidueProperties.aminoBackgroundFrequencies
                .containsKey(symbol))
        {

          Double prob;
          Float bfreq;
          Double llr;
          prob = hmm.getMatchEmissionProbability(alignPos, symbol);
          bfreq = ResidueProperties.aminoBackgroundFrequencies.get(symbol);
          llr = Math.log(prob / bfreq);
          raw.get(rawPos).add(llr);
          String output;
          output = String.format("%.1f", llr);
          if ("-0.0".equals(output))
          {
            output = "0.0";
          }
          if (binned.containsKey(output))
          {
            double prev = binned.get(output);
            prev += (SCALE / count);
            binned.put(output, prev);

          }
          else
          {
            binned.put(output, SCALE / count);
          }
        }
      }
    }
  }


  /**
   * Reads in the sequence data from a Stockholm file.
   * 
   * @param source
   * @throws IOException
   */
  public void readStockholm(FileParse source) throws IOException
  {
    StockholmFile file = new StockholmFile(source);
    file.parse();
    sequences = file.getSeqs();
  }

  /**
   * Reads in the HMM data from a HMMer file.
   * 
   * @param source
   * @throws IOException
   */
  public void readHMM(FileParse source) throws IOException
  {

    HMMFile file = new HMMFile(source);
    file.parse();
    hmm = file.getHMM();

  }

  /**
   * Exports both the binned and raw data into separate files.
   * 
   * @param location
   * @throws FileNotFoundException
   */
  public void exportData(String location) throws FileNotFoundException
  {
    PrintWriter writerBin = new PrintWriter(new File(location + BINNED));
    for (Map.Entry<String, Double> entry : binned.entrySet())
    {
      writerBin.println(entry.getKey() + "," + entry.getValue());
    }
    writerBin.close();

    PrintWriter writerRaw = new PrintWriter(new File(location + RAW));

    StringBuilder identifier = new StringBuilder();

    for (int i = 1; i < raw.size() + 1; i++)
    {
      identifier.append("Fam " + i + ",");
    }

    writerRaw.println(identifier);

    boolean rowIsEmpty = false;
    int row = 0;
    while (!rowIsEmpty)
    {
      rowIsEmpty = true;
      StringBuilder string = new StringBuilder();
      for (int column = 0; column < raw.size(); column++)
      {
        if (raw.get(column).size() <= row)
        {
          string.append("EMPTY,");
        }
        else
        {
          string.append(raw.get(column).get(row) + ",");
          rowIsEmpty = false;
        }
      }
      row++;
      writerRaw.println(string);
    }
    writerRaw.close();

  }

  /**
   * Prints the specified family on the console.
   * 
   * @param index
   * @throws IOException
   */
  public void printFam(int index) throws IOException
  {
    BufferedReader br = new BufferedReader(new FileReader(FAMILIES));

    moveToFile(index, br);

    String line = br.readLine();

    while (!"//".equals(line))
    {
      System.out.println(line);
      line = br.readLine();
    }
    System.out.println(line);
    br.close();

  }

  /**
   * Prints the specified HMM on the console.
   * 
   * @param index
   * @throws IOException
   */
  public void printHMM(int index) throws IOException
  {
    BufferedReader br = new BufferedReader(new FileReader(HMMS));

    moveToFile(index, br);

    String line = br.readLine();

    while (!"//".equals(line))
    {
      System.out.println(line);
      line = br.readLine();
    }
    System.out.println(line);
    br.close();

  }

  /**
   * Prints the specified family to a .sto file in the current directory.
   * 
   * @param index
   * @throws IOException
   */
  public void printFamToFile(int index) throws IOException
  {
    String name;

    BufferedReader nameFinder = new BufferedReader(
            new FileReader(FAMILIES));

    moveToFile(index, nameFinder);

    nameFinder.readLine();

    Scanner scanner = new Scanner(nameFinder.readLine());
    scanner.next();
    scanner.next();
    name = scanner.next();
    scanner.close();

    BufferedReader br = new BufferedReader(new FileReader(FAMILIES));

    moveToFile(index, br);

    String line = br.readLine();
    PrintWriter writer = new PrintWriter(
            currentFolder + "/" + name + ".sto");
    while (!"//".equals(line))
    {
      writer.println(line);
      line = br.readLine();
    }
    writer.println(line);
    writer.close();
    br.close();

  }

  /**
   * Prints the specified family to a .hmm file in the current directory.
   * 
   * @param index
   * @throws IOException
   */
  public void printHMMToFile(int index) throws IOException
  {

    String name;

    BufferedReader nameFinder = new BufferedReader(new FileReader(HMMS));

    moveToFile(index, nameFinder);

    nameFinder.readLine();

    Scanner scanner = new Scanner(nameFinder.readLine());
    name = scanner.next();
    name = scanner.next();
    scanner.close();

    BufferedReader br = new BufferedReader(new FileReader(HMMS));

    moveToFile(index, br);

    String line = br.readLine();

    PrintWriter writer = new PrintWriter(
            currentFolder + "/" + name + ".hmm");
    while (!"//".equals(line))
    {
      writer.println(line);
      line = br.readLine();
    }
    writer.println(line);
    writer.close();
    br.close();

  }
  
  /**
   * Clears all raw, binned and current position data in the current directory.
   * 
   * @throws FileNotFoundException
   */
  public void clear() throws FileNotFoundException
  {
    PrintWriter pos = new PrintWriter(
            currentFolder + "/CurrentPosition.txt");
    pos.println("0");
    
    PrintWriter raw = new PrintWriter(currentFolder + RAW);
    
    PrintWriter bin = new PrintWriter(currentFolder + BINNED);
    
    pos.close();
    bin.close();
    raw.close();
  }

}
